#include "qemu/osdep.h"
#include "migration/channel-savevm-async.h"
#include "migration/migration.h"
#include "migration/savevm.h"
#include "migration/snapshot.h"
#include "migration/global_state.h"
#include "migration/ram.h"
#include "migration/qemu-file.h"
#include "sysemu/sysemu.h"
#include "sysemu/runstate.h"
#include "block/block.h"
#include "sysemu/block-backend.h"
#include "qapi/error.h"
#include "qapi/qmp/qerror.h"
#include "qapi/qmp/qdict.h"
#include "qapi/qapi-commands-migration.h"
#include "qapi/qapi-commands-misc.h"
#include "qapi/qapi-commands-block.h"
#include "qemu/cutils.h"
#include "qemu/timer.h"
#include "qemu/main-loop.h"
#include "qemu/rcu.h"
#include "qemu/yank.h"

/* #define DEBUG_SAVEVM_STATE */

#ifdef DEBUG_SAVEVM_STATE
#define DPRINTF(fmt, ...) \
    do { printf("savevm-async: " fmt, ## __VA_ARGS__); } while (0)
#else
#define DPRINTF(fmt, ...) \
    do { } while (0)
#endif

enum {
    SAVE_STATE_DONE,
    SAVE_STATE_ERROR,
    SAVE_STATE_ACTIVE,
    SAVE_STATE_COMPLETED,
    SAVE_STATE_CANCELLED
};


static struct SnapshotState {
    BlockBackend *target;
    size_t bs_pos;
    int state;
    Error *error;
    Error *blocker;
    int saved_vm_running;
    QEMUFile *file;
    int64_t total_time;
    QEMUBH *finalize_bh;
    Coroutine *co;
    QemuCoSleep target_close_wait;
} snap_state;

static bool savevm_aborted(void)
{
    return snap_state.state == SAVE_STATE_CANCELLED ||
        snap_state.state == SAVE_STATE_ERROR;
}

SaveVMInfo *qmp_query_savevm(Error **errp)
{
    SaveVMInfo *info = g_malloc0(sizeof(*info));
    struct SnapshotState *s = &snap_state;

    if (s->state != SAVE_STATE_DONE) {
        info->has_bytes = true;
        info->bytes = s->bs_pos;
        switch (s->state) {
        case SAVE_STATE_ERROR:
            info->has_status = true;
            info->status = g_strdup("failed");
            info->has_total_time = true;
            info->total_time = s->total_time;
            if (s->error) {
                info->has_error = true;
                info->error = g_strdup(error_get_pretty(s->error));
            }
            break;
        case SAVE_STATE_ACTIVE:
            info->has_status = true;
            info->status = g_strdup("active");
            info->has_total_time = true;
            info->total_time = qemu_clock_get_ms(QEMU_CLOCK_REALTIME)
                - s->total_time;
            break;
        case SAVE_STATE_COMPLETED:
            info->has_status = true;
            info->status = g_strdup("completed");
            info->has_total_time = true;
            info->total_time = s->total_time;
            break;
        }
    }

    return info;
}

static int save_snapshot_cleanup(void)
{
    int ret = 0;

    DPRINTF("save_snapshot_cleanup\n");

    snap_state.total_time = qemu_clock_get_ms(QEMU_CLOCK_REALTIME) -
        snap_state.total_time;

    if (snap_state.file) {
        ret = qemu_fclose(snap_state.file);
        snap_state.file = NULL;
    }

    if (snap_state.target) {
        if (!savevm_aborted()) {
            /* try to truncate, but ignore errors (will fail on block devices).
            * note1: bdrv_read() need whole blocks, so we need to round up
            * note2: PVE requires 1024 (BDRV_SECTOR_SIZE*2) alignment
            */
            size_t size = QEMU_ALIGN_UP(snap_state.bs_pos, BDRV_SECTOR_SIZE*2);
            blk_truncate(snap_state.target, size, false, PREALLOC_MODE_OFF, 0, NULL);
        }
        blk_op_unblock_all(snap_state.target, snap_state.blocker);
        error_free(snap_state.blocker);
        snap_state.blocker = NULL;
        blk_unref(snap_state.target);
        snap_state.target = NULL;

        qemu_co_sleep_wake(&snap_state.target_close_wait);
    }

    return ret;
}

static void save_snapshot_error(const char *fmt, ...)
{
    va_list ap;
    char *msg;

    va_start(ap, fmt);
    msg = g_strdup_vprintf(fmt, ap);
    va_end(ap);

    DPRINTF("save_snapshot_error: %s\n", msg);

    if (!snap_state.error) {
        error_set(&snap_state.error, ERROR_CLASS_GENERIC_ERROR, "%s", msg);
    }

    g_free (msg);

    snap_state.state = SAVE_STATE_ERROR;
}

static void process_savevm_finalize(void *opaque)
{
    int ret;
    AioContext *iohandler_ctx = iohandler_get_aio_context();
    MigrationState *ms = migrate_get_current();

    bool aborted = savevm_aborted();

#ifdef DEBUG_SAVEVM_STATE
    int64_t start_time = qemu_clock_get_ms(QEMU_CLOCK_REALTIME);
#endif

    qemu_bh_delete(snap_state.finalize_bh);
    snap_state.finalize_bh = NULL;
    snap_state.co = NULL;

    /* We need to own the target bdrv's context for the following functions,
     * so move it back. It can stay in the main context and live out its live
     * there, since we're done with it after this method ends anyway.
     */
    aio_context_acquire(iohandler_ctx);
    blk_set_aio_context(snap_state.target, qemu_get_aio_context(), NULL);
    aio_context_release(iohandler_ctx);

    ret = vm_stop_force_state(RUN_STATE_FINISH_MIGRATE);
    if (ret < 0) {
        save_snapshot_error("vm_stop_force_state error %d", ret);
    }

    if (!aborted) {
        /* skip state saving if we aborted, snapshot will be invalid anyway */
        (void)qemu_savevm_state_complete_precopy(snap_state.file, false, false);
        ret = qemu_file_get_error(snap_state.file);
        if (ret < 0) {
            save_snapshot_error("qemu_savevm_state_complete_precopy error %d", ret);
        }
    }

    DPRINTF("state saving complete\n");
    DPRINTF("timing: process_savevm_finalize (state saving) took %ld ms\n",
        qemu_clock_get_ms(QEMU_CLOCK_REALTIME) - start_time);

    /* clear migration state */
    migrate_set_state(&ms->state, MIGRATION_STATUS_SETUP,
        ret || aborted ? MIGRATION_STATUS_FAILED : MIGRATION_STATUS_COMPLETED);
    ms->to_dst_file = NULL;

    qemu_savevm_state_cleanup();

    ret = save_snapshot_cleanup();
    if (ret < 0) {
        save_snapshot_error("save_snapshot_cleanup error %d", ret);
    } else if (snap_state.state == SAVE_STATE_ACTIVE) {
        snap_state.state = SAVE_STATE_COMPLETED;
    } else if (aborted) {
        /*
         * If there was an error, there's no need to set a new one here.
         * If the snapshot was canceled, leave setting the state to
         * qmp_savevm_end(), which is waked by save_snapshot_cleanup().
         */
    } else {
        save_snapshot_error("process_savevm_cleanup: invalid state: %d",
                            snap_state.state);
    }
    if (snap_state.saved_vm_running) {
        vm_start();
        snap_state.saved_vm_running = false;
    }

    DPRINTF("timing: process_savevm_finalize (full) took %ld ms\n",
        qemu_clock_get_ms(QEMU_CLOCK_REALTIME) - start_time);
}

static void coroutine_fn process_savevm_co(void *opaque)
{
    int ret;
    int64_t maxlen;
    BdrvNextIterator it;
    BlockDriverState *bs = NULL;

#ifdef DEBUG_SAVEVM_STATE
    int64_t start_time = qemu_clock_get_ms(QEMU_CLOCK_REALTIME);
#endif

    ret = qemu_file_get_error(snap_state.file);
    if (ret < 0) {
        save_snapshot_error("qemu_savevm_state_setup failed");
        return;
    }

    while (snap_state.state == SAVE_STATE_ACTIVE) {
        uint64_t pending_size, pend_precopy, pend_compatible, pend_postcopy;

        /* pending is expected to be called without iothread lock */
        qemu_mutex_unlock_iothread();
        qemu_savevm_state_pending(snap_state.file, 0, &pend_precopy, &pend_compatible, &pend_postcopy);
        qemu_mutex_lock_iothread();

        pending_size = pend_precopy + pend_compatible + pend_postcopy;

        /*
         * A guest reaching this cutoff is dirtying lots of RAM. It should be
         * large enough so that the guest can't dirty this much between the
         * check and the guest actually being stopped, but it should be small
         * enough to avoid long downtimes for non-hibernation snapshots.
         */
        maxlen = blk_getlength(snap_state.target) - 100*1024*1024;

        /* Note that there is no progress for pend_postcopy when iterating */
        if (pending_size - pend_postcopy > 400000 && snap_state.bs_pos + pending_size < maxlen) {
            ret = qemu_savevm_state_iterate(snap_state.file, false);
            if (ret < 0) {
                save_snapshot_error("qemu_savevm_state_iterate error %d", ret);
                break;
            }
            DPRINTF("savevm iterate pending size %lu ret %d\n", pending_size, ret);
        } else {
            qemu_system_wakeup_request(QEMU_WAKEUP_REASON_OTHER, NULL);
            ret = global_state_store();
            if (ret) {
                save_snapshot_error("global_state_store error %d", ret);
                break;
            }

            DPRINTF("savevm iterate complete\n");
            break;
        }
    }

    DPRINTF("timing: process_savevm_co took %ld ms\n",
        qemu_clock_get_ms(QEMU_CLOCK_REALTIME) - start_time);

#ifdef DEBUG_SAVEVM_STATE
    int64_t start_time_flush = qemu_clock_get_ms(QEMU_CLOCK_REALTIME);
#endif
    /* If a drive runs in an IOThread we can flush it async, and only
     * need to sync-flush whatever IO happens between now and
     * vm_stop_force_state. bdrv_next can only be called from main AioContext,
     * so move there now and after every flush.
     */
    aio_co_reschedule_self(qemu_get_aio_context());
    for (bs = bdrv_first(&it); bs; bs = bdrv_next(&it)) {
        /* target has BDRV_O_NO_FLUSH, no sense calling bdrv_flush on it */
        if (bs == blk_bs(snap_state.target)) {
            continue;
        }

        AioContext *bs_ctx = bdrv_get_aio_context(bs);
        if (bs_ctx != qemu_get_aio_context()) {
            DPRINTF("savevm: async flushing drive %s\n", bs->filename);
            aio_co_reschedule_self(bs_ctx);
            bdrv_flush(bs);
            aio_co_reschedule_self(qemu_get_aio_context());
        }
    }

    DPRINTF("timing: async flushing took %ld ms\n",
        qemu_clock_get_ms(QEMU_CLOCK_REALTIME) - start_time_flush);

    qemu_bh_schedule(snap_state.finalize_bh);
}

void qmp_savevm_start(bool has_statefile, const char *statefile, Error **errp)
{
    Error *local_err = NULL;
    MigrationState *ms = migrate_get_current();
    AioContext *iohandler_ctx = iohandler_get_aio_context();

    int bdrv_oflags = BDRV_O_RDWR | BDRV_O_RESIZE | BDRV_O_NO_FLUSH;

    if (snap_state.state != SAVE_STATE_DONE) {
        error_set(errp, ERROR_CLASS_GENERIC_ERROR,
                  "VM snapshot already started\n");
        return;
    }

    if (migration_is_running(ms->state)) {
        error_set(errp, ERROR_CLASS_GENERIC_ERROR, QERR_MIGRATION_ACTIVE);
        return;
    }

    if (migrate_use_block()) {
        error_set(errp, ERROR_CLASS_GENERIC_ERROR,
                  "Block migration and snapshots are incompatible");
        return;
    }

    /* initialize snapshot info */
    snap_state.saved_vm_running = runstate_is_running();
    snap_state.bs_pos = 0;
    snap_state.total_time = qemu_clock_get_ms(QEMU_CLOCK_REALTIME);
    snap_state.blocker = NULL;
    snap_state.target_close_wait = (QemuCoSleep){ .to_wake = NULL };

    if (snap_state.error) {
        error_free(snap_state.error);
        snap_state.error = NULL;
    }

    if (!has_statefile) {
        vm_stop(RUN_STATE_SAVE_VM);
        snap_state.state = SAVE_STATE_COMPLETED;
        return;
    }

    if (qemu_savevm_state_blocked(errp)) {
        return;
    }

    /* Open the image */
    QDict *options = NULL;
    options = qdict_new();
    qdict_put_str(options, "driver", "raw");
    snap_state.target = blk_new_open(statefile, NULL, options, bdrv_oflags, &local_err);
    if (!snap_state.target) {
        error_set(errp, ERROR_CLASS_GENERIC_ERROR, "failed to open '%s'", statefile);
        goto restart;
    }

    QIOChannel *ioc = QIO_CHANNEL(qio_channel_savevm_async_new(snap_state.target,
                                                               &snap_state.bs_pos));
    snap_state.file = qemu_file_new_output_sized(ioc, 4 * 1024 * 1024);

    if (!snap_state.file) {
        error_set(errp, ERROR_CLASS_GENERIC_ERROR, "failed to open '%s'", statefile);
        goto restart;
    }

    /*
     * qemu_savevm_* paths use migration code and expect a migration state.
     * State is cleared in process_savevm_co, but has to be initialized
     * here (blocking main thread, from QMP) to avoid race conditions.
     */
    migrate_init(ms);
    memset(&ram_counters, 0, sizeof(ram_counters));
    ms->to_dst_file = snap_state.file;

    error_setg(&snap_state.blocker, "block device is in use by savevm");
    blk_op_block_all(snap_state.target, snap_state.blocker);

    snap_state.state = SAVE_STATE_ACTIVE;
    snap_state.finalize_bh = qemu_bh_new(process_savevm_finalize, &snap_state);
    snap_state.co = qemu_coroutine_create(&process_savevm_co, NULL);
    qemu_mutex_unlock_iothread();
    qemu_savevm_state_header(snap_state.file);
    qemu_savevm_state_setup(snap_state.file);
    qemu_mutex_lock_iothread();

    /* Async processing from here on out happens in iohandler context, so let
     * the target bdrv have its home there.
     */
    blk_set_aio_context(snap_state.target, iohandler_ctx, &local_err);

    aio_co_schedule(iohandler_ctx, snap_state.co);

    return;

restart:

    save_snapshot_error("setup failed");

    if (snap_state.saved_vm_running) {
        vm_start();
        snap_state.saved_vm_running = false;
    }
}

void coroutine_fn qmp_savevm_end(Error **errp)
{
    int64_t timeout;

    if (snap_state.state == SAVE_STATE_DONE) {
        error_set(errp, ERROR_CLASS_GENERIC_ERROR,
                  "VM snapshot not started\n");
        return;
    }

    if (snap_state.state == SAVE_STATE_ACTIVE) {
        snap_state.state = SAVE_STATE_CANCELLED;
        goto wait_for_close;
    }

    if (snap_state.saved_vm_running) {
        vm_start();
        snap_state.saved_vm_running = false;
    }

    snap_state.state = SAVE_STATE_DONE;

wait_for_close:
    if (!snap_state.target) {
        DPRINTF("savevm-end: no target file open\n");
        return;
    }

    /* wait until cleanup is done before returning, this ensures that after this
     * call exits the statefile will be closed and can be removed immediately */
    DPRINTF("savevm-end: waiting for cleanup\n");
    timeout = 30L * 1000 * 1000 * 1000;
    qemu_co_sleep_ns_wakeable(&snap_state.target_close_wait,
                              QEMU_CLOCK_REALTIME, timeout);
    if (snap_state.target) {
        save_snapshot_error("timeout waiting for target file close in "
                            "qmp_savevm_end");
        /* we cannot assume the snapshot finished in this case, so leave the
         * state alone - caller has to figure something out */
        return;
    }

    // File closed and no other error, so ensure next snapshot can be started.
    if (snap_state.state != SAVE_STATE_ERROR) {
        snap_state.state = SAVE_STATE_DONE;
    }

    DPRINTF("savevm-end: cleanup done\n");
}

// FIXME: Deprecated
void qmp_snapshot_drive(const char *device, const char *name, Error **errp)
{
    // Compatibility to older qemu-server.
    qmp_blockdev_snapshot_internal_sync(device, name, errp);
}

// FIXME: Deprecated
void qmp_delete_drive_snapshot(const char *device, const char *name,
                               Error **errp)
{
    // Compatibility to older qemu-server.
    (void)qmp_blockdev_snapshot_delete_internal_sync(device, false, NULL,
                                                     true, name, errp);
}

int load_snapshot_from_blockdev(const char *filename, Error **errp)
{
    BlockBackend *be;
    Error *local_err = NULL;
    Error *blocker = NULL;

    QEMUFile *f;
    size_t bs_pos = 0;
    int ret = -EINVAL;

    be = blk_new_open(filename, NULL, NULL, 0, &local_err);

    if (!be) {
        error_setg(errp, "Could not open VM state file");
        goto the_end;
    }

    error_setg(&blocker, "block device is in use by load state");
    blk_op_block_all(be, blocker);

    /* restore the VM state */
    f = qemu_file_new_input_sized(QIO_CHANNEL(qio_channel_savevm_async_new(be, &bs_pos)),
                                  4 * 1024 * 1024);
    if (!f) {
        error_setg(errp, "Could not open VM state file");
        goto the_end;
    }

    qemu_system_reset(SHUTDOWN_CAUSE_NONE);
    ret = qemu_loadvm_state(f);

    /* dirty bitmap migration has a special case we need to trigger manually */
    dirty_bitmap_mig_before_vm_start();

    qemu_fclose(f);

    /* state_destroy assumes a real migration which would have added a yank */
    yank_register_instance(MIGRATION_YANK_INSTANCE, &error_abort);

    migration_incoming_state_destroy();
    if (ret < 0) {
        error_setg_errno(errp, -ret, "Error while loading VM state");
        goto the_end;
    }

    ret = 0;

 the_end:
    if (be) {
        blk_op_unblock_all(be, blocker);
        error_free(blocker);
        blk_unref(be);
    }
    return ret;
}
